/***************************************************************************
 * Copyright 2012 Kieker Project (http://kieker-monitoring.net)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 ***************************************************************************/

package kieker.webgui.common.util;

import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.URLClassLoader;
import java.nio.channels.ByteChannel;
import java.nio.channels.FileChannel;
import java.security.AccessController;
import java.security.PrivilegedAction;
import java.util.ArrayList;
import java.util.Collection;
import java.util.List;

import kieker.analysis.AnalysisController;
import kieker.analysis.model.analysisMetaModel.MIAnalysisMetaModelFactory;
import kieker.analysis.model.analysisMetaModel.MIDependency;
import kieker.analysis.model.analysisMetaModel.MIProject;
import kieker.analysis.model.analysisMetaModel.impl.MAnalysisMetaModelFactory;
import kieker.common.logging.Log;
import kieker.common.logging.LogFactory;
import kieker.webgui.common.ClassAndMethodContainer;
import kieker.webgui.common.exception.LibraryAlreadyExistingException;
import kieker.webgui.common.exception.NewerProjectException;
import kieker.webgui.common.exception.ProjectAlreadyExistingException;
import kieker.webgui.common.exception.ProjectNotExistingException;

import org.primefaces.model.UploadedFile;

/**
 * This is a singleton class for the access to the file system. It makes sure that the necessary directories for the execution of the application exist. <b>Do
 * not</b> remove directories created from this manager during runtime! Directories are created during first access to the class.<br>
 * This class <b>does not</b> use any kind of synchronization to handle the access to the projects.
 * 
 * @author Nils Christian Ehmke
 */
public final class FSManager {
	/**
	 * This is the log object used to log messages, warnings etc.
	 */
	private static final Log LOG = LogFactory.getLog(FSManager.class);
	/**
	 * This is the extension of the KAX-files.
	 */
	private static final String KAX_EXTENSION = "kax";
	/**
	 * This is the extension of the library-files.
	 */
	private static final String LIB_EXTENSION = "jar";
	/**
	 * This is the name of the directory for the libraries.
	 */
	private static final String LIB_DIRECTORY = "lib";
	/**
	 * This is the name of the root-directory.
	 */
	private static final String ROOT_DIRECTORY = "data";
	/**
	 * The library for kieker which is contained in the war-file as a resource.
	 */
	private static final String KIEKER_LIB = "kieker-1.6-SNAPSHOT_emf.jar";
	/**
	 * This is the buffer (in bytes) used to copy and upload files.
	 */
	private static final int BUF_SIZE_BYTES = 1024 * 1024;
	/**
	 * This is the singleton instance of this class.
	 */
	private static final FSManager INSTANCE = new FSManager();
	/**
	 * This is the factory we will use to create the elements for the projects.
	 */
	private final MIAnalysisMetaModelFactory factory = new MAnalysisMetaModelFactory();

	/**
	 * Creates a new instance of this class.
	 */
	private FSManager() {
		this.initialize();
	}

	/**
	 * This method initializes the class. Normally this method should be called via the @PostConstruction annotation, but for unknown reason it does only work for
	 * beans.
	 */
	private void initialize() {
		// Check for our root-directory and create it if necessary
		final File rootDir = new File(FSManager.ROOT_DIRECTORY);
		if (!rootDir.exists()) {
			final boolean result = rootDir.mkdir();
			if (!result) {
				FSManager.LOG.error("Could not create root directory.");
			}
		}
	}

	/**
	 * Delivers the singleton instance of this class.
	 * 
	 * @return The only instance of {@link FSManager}.
	 */
	public static FSManager getInstance() {
		return FSManager.INSTANCE;
	}

	/**
	 * This method adds a new project to the application. It creates an empty, but nevertheless valid kax-file to the file system. If the method fails due to an
	 * {@link IOException}, it will make sure that the project-directories will be removed as if the method would never have been called.
	 * 
	 * @param projectName
	 *            The name of the new project.
	 * @throws ProjectAlreadyExistingException
	 *             If a project with the same name exists already.
	 * @throws IOException
	 *             If something went wrong during the creation of the project.
	 */
	public void addProject(final String projectName) throws ProjectAlreadyExistingException, IOException {
		// Assemble all necessary paths and files for the given project
		final File projectDir = this.assembleProjectDir(projectName);
		final File projectFile = this.assembleKaxFile(projectName);
		final File libDir = this.assembleLibDir(projectName);

		// We need an "empty" project in order to save it.
		final MIProject emptyProject = this.factory.createProject();

		// Make sure that the project doesn't exist already
		if (projectDir.exists()) {
			throw new ProjectAlreadyExistingException("The project with the name '" + projectName + "' exists already.");
		}

		try {
			// Try to create the directories
			if (projectDir.mkdir() && libDir.mkdir()) {
				// Try to save the file
				AnalysisController.saveToFile(projectFile, emptyProject);
			} else {
				// The directories could not be created
				throw new IOException("Project-Directories could not be created.");
			}
		} catch (final IOException ex) {
			// Something went wrong. Remove the directories and files!
			final boolean libDirDeleted = libDir.delete();
			// Keep in mind that the potential remains of the file have to be deleted before the directory.
			final boolean projectFileDeleted = projectFile.delete();
			final boolean projectDeleted = projectDir.delete();

			// The following part is only necessary to calm FindBugs...
			@SuppressWarnings("unused")
			final boolean deleteResults = libDirDeleted && projectFileDeleted && projectDeleted;

			// Rethrow the exception in order to inform the caller of this method
			throw new IOException("An IO Exception occured while saving the project.", ex);
		}
	}

	/**
	 * This method makes a copy of a project and saves it under another name. If the method fails due to an {@link IOException}, it will make sure that the
	 * project-directories of the destination-project will be removed as if the method would never have been called.
	 * 
	 * @param originalProjectName
	 *            The name of the source project.
	 * @param newProjectName
	 *            The name of the target project.
	 * @throws ProjectNotExistingException
	 *             If a project with the given (source) name doesn't exist.
	 * @throws ProjectAlreadyExistingException
	 *             If a project with the same (target) name exists already.
	 * @throws IOException
	 *             If something went wrong during the creation of the target-project or during the loading of the source-project.
	 */
	public void copyProject(final String originalProjectName, final String newProjectName) throws ProjectNotExistingException, ProjectAlreadyExistingException,
			IOException {
		// Get the necessary paths
		final File dstProjDir = this.assembleProjectDir(newProjectName);
		final File srcLibDir = this.assembleLibDir(originalProjectName);
		final File dstLibDir = this.assembleLibDir(newProjectName);

		// Check whether the project exists already!
		if (this.projectExists(newProjectName)) {
			throw new ProjectAlreadyExistingException("A project with the name '" + newProjectName + "' exists already.");
		}

		final File srcKaxFile = this.assembleKaxFile(originalProjectName);
		final File dstKaxFile = this.assembleKaxFile(newProjectName);

		try {
			if (dstProjDir.mkdir() && dstLibDir.mkdir()) {
				// Copy the kax file
				this.copyFile(srcKaxFile, dstKaxFile);

				// Copy the libs
				for (final File lib : srcLibDir.listFiles()) {
					final File dstLibFile = new File(dstLibDir, lib.getName());
					this.copyFile(lib, dstLibFile);
				}
			} else {
				throw new IOException("Could not create directories.");
			}
		} catch (final IOException ex) {
			// Something went wrong. Remove the directories and files!
			final boolean libDirDeleted = dstLibDir.delete();
			// Keep in mind that the potential remains of the file have to be deleted before the directory.
			final boolean projectFileDeleted = dstKaxFile.delete();
			final boolean projectDeleted = dstProjDir.delete();

			// The following part is only necessary to calm FindBugs...
			@SuppressWarnings("unused")
			final boolean deleteResults = libDirDeleted && projectFileDeleted && projectDeleted;

			// Rethrow the exception in order to inform the caller of this method
			throw new IOException("An IO Exception occured while copying the project.", ex);
		}
	}

	/**
	 * This method loads the kax-file for the given project name and delivers an initializes instance of {@link MIProject}.
	 * 
	 * @param projectName
	 *            The name of the project to be loaded.
	 * @return The model instance as defined by the corresponding kax-file.
	 * @throws ProjectNotExistingException
	 *             If a project with the given (source) name doesn't exist.
	 * @throws IOException
	 *             If something went wrong during the opening of the project.
	 */
	public MIProject openProject(final String projectName) throws ProjectNotExistingException, IOException {
		if (projectName == null) {
			throw new IOException("Project is null");
		}

		try {
			// Load the project
			return AnalysisController.loadFromFile(this.assembleKaxFile(projectName).getAbsoluteFile());
		} catch (final FileNotFoundException ex) {
			throw new ProjectNotExistingException("A project with the name '" + projectName + "' does not exist.", ex);
		}
	}

	/**
	 * This method loads the kax-file for the given project name and delivers an initializes instance of {@link MIProject} - but instead of using the "normal" class
	 * loader, it uses the methods and classes stored in the given instance of {@link ClassAndMethodContainer}. This means that this method <b>does</b> return an
	 * instance of {@link MIProject}, but the one defined in the container. This is also the reason why this method has to return an {@link Object}-instance.
	 * 
	 * @param projectName
	 *            The name of the project to be loaded.
	 * @param classAndMethodContainer
	 *            The container, which will be used to load the project instance.
	 * @return The model instance as defined by the corresponding kax-file.
	 * @throws ProjectNotExistingException
	 *             If a project with the given (source) name doesn't exist.
	 * @throws IOException
	 *             If something went wrong during the opening of the project. This can also mean that the given {@link ClassAndMethodContainer} is somehow invalid.
	 */
	public Object openProject(final String projectName, final ClassAndMethodContainer classAndMethodContainer) throws ProjectNotExistingException, IOException {
		if (projectName == null) {
			throw new IOException("Project is null");
		}

		// Load the project
		final Object project = ClassAndMethodContainer.invokeClassMethod(classAndMethodContainer.getAnalysisControllerLoadFromFile(), null,
				this.assembleKaxFile(projectName));

		if (project == null) {
			throw new IOException("Project could not be loaded.");
		} else {
			return project;
		}
	}

	/**
	 * This method tries to save the given model instance for the given project. The given time stamp will be compared (if the corresponding flag says so) with the
	 * current time stamp of the project. If the project on the file system has been modified in the meantime, a {@link NewerProjectException} will be thrown. If
	 * something goes wrong during the storage, it is <b>not</b> guaranteed that the resulting file will be valid.
	 * 
	 * @param projectName
	 *            The name of the project which has to be saved.
	 * @param project
	 *            The model instance to be stored in the corresponding kax-file.
	 * @param timeStamp
	 *            The time stamp which has to be compared with the "real" time stamp of the project.
	 * @param overwriteNewerProject
	 *            Determines whether a newer project file will be overwritten without further warning or not-
	 * @throws ProjectNotExistingException
	 *             If a project with the given name does not exist.
	 * @throws IOException
	 *             If something went wrong during the storage of the model instance.
	 * @throws NewerProjectException
	 *             If the project on the file system is newer and the overwriteNewerProject-flag has not been set.
	 */
	public void saveProject(final String projectName, final MIProject project, final long timeStamp, final boolean overwriteNewerProject) throws
			ProjectNotExistingException, IOException, NewerProjectException {
		// Check whether the project exists
		if (!this.projectExists(projectName)) {
			throw new ProjectNotExistingException("A project with the name '" + projectName + "' does not exist.");
		}

		// Check for a newer version first
		final long currTimeStamp = this.getCurrTimeStamp(projectName);
		if (!overwriteNewerProject && (currTimeStamp > timeStamp)) {
			throw new NewerProjectException("The project with the name '" + projectName + "' has a newer version on the FS.");
		}

		// Try to save it.
		AnalysisController.saveToFile(this.assembleKaxFile(projectName), project);
	}

	/**
	 * Delivers the current time stamp of the given project.
	 * 
	 * @param projectName
	 *            The name of the project whose time stamp will be delivered.
	 * @return The current time stamp.
	 * @throws ProjectNotExistingException
	 *             If a project with the given name does not exist.
	 */
	public long getCurrTimeStamp(final String projectName) throws ProjectNotExistingException {
		// Check whether the project exists
		if (!this.projectExists(projectName)) {
			throw new ProjectNotExistingException("A project with the name '" + projectName + "' does not exist.");
		}
		return this.assembleKaxFile(projectName).lastModified();
	}

	/**
	 * This method tries to upload a dependency to the given project.
	 * 
	 * @param file
	 *            The file to be uploaded to the project.
	 * @param projectName
	 *            The name of the project.
	 * @throws ProjectNotExistingException
	 *             If a project with the given name does not exist.
	 * @throws IOException
	 *             If something went wrong during the uploading.
	 * @throws LibraryAlreadyExistingException
	 *             If a library with the same name exists already.
	 */
	public void uploadLibrary(final UploadedFile file, final String projectName) throws ProjectNotExistingException, IOException, LibraryAlreadyExistingException {
		// Check whether the project exists
		if (!this.projectExists(projectName)) {
			throw new ProjectNotExistingException("A project with the name '" + projectName + "' does not exist.");
		}

		// Prepare the files
		final File libDir = this.assembleLibDir(projectName);
		final File dstFile = new File(libDir, file.getFileName());

		// Now copy the file - if it doesn't already exist
		if (dstFile.exists()) {
			throw new LibraryAlreadyExistingException("The library with the name '" + file.getFileName() + "' exists already.");
		}

		BufferedInputStream in = null;
		BufferedOutputStream out = null;
		IOException occException = null;
		try {
			// Get the streams.
			in = new BufferedInputStream(file.getInputstream());
			out = new BufferedOutputStream(new FileOutputStream(dstFile));
			final byte[] buf = new byte[FSManager.BUF_SIZE_BYTES];
			int count;

			// Transfer the file.
			count = in.read(buf);
			while (count != -1) {
				out.write(buf, 0, count);
				count = in.read(buf);
			}
		} finally {
			// Try to make sure that the streams will be closed.
			try {
				if (in != null) {
					in.close();
				}
			} catch (final IOException ex) {
				occException = ex;
			}

			try {
				if (out != null) {
					out.close();
				}
			} catch (final IOException ex) {
				occException = ex;
			}
		}
		// If something went wrong, rethrow the exception
		if (occException != null) {
			throw occException;
		}
	}

	/**
	 * This method delivers a class loader containing the currently available libraries of the given project.
	 * 
	 * @param projectName
	 *            The name of the project.
	 * @return A class loader for the given project.
	 * @throws ProjectNotExistingException
	 *             If a project with the given name does not exist.
	 * @throws IOException
	 *             If something went wrong during the initialization of the class loader.
	 */
	public ClassLoader getClassLoader(final String projectName) throws ProjectNotExistingException, IOException {
		// Check whether the project exists
		if (!this.projectExists(projectName)) {
			throw new ProjectNotExistingException("A project with the name '" + projectName + "' does not exist.");
		}

		final List<URL> libs = new ArrayList<URL>();

		// Collect all libraries of the project

		// Run through the libs and put them into our list.
		final File libDir = new File(FSManager.ROOT_DIRECTORY + File.separator + projectName + File.separator + FSManager.LIB_DIRECTORY);
		final File[] files = libDir.listFiles();
		if (files != null) {
			for (final File file : files) {
				if (file.getName().endsWith("." + FSManager.LIB_EXTENSION)) {
					try {
						libs.add(file.toURL());
					} catch (final MalformedURLException ex) {
						ex.printStackTrace();
					}
				}
			}
		}

		// Add the kieker lib!
		libs.add(this.getKiekerURL());

		// Now assemble the URL class loader
		final PrivilegedClassLoaderAction action = new PrivilegedClassLoaderAction(libs);
		return AccessController.doPrivileged(action);
	}

	/**
	 * This method lists all available libraries of the given project.
	 * 
	 * @param projectName
	 *            The name of the project whose libraries have to be delivered.
	 * @return A list containing all available library-names of the project.
	 * @throws ProjectNotExistingException
	 *             If a project with the given name does not exist.
	 */
	public List<String> listAllLibraries(final String projectName) throws ProjectNotExistingException {
		// Check whether the project exists
		if (!this.projectExists(projectName)) {
			throw new ProjectNotExistingException("A project with the name '" + projectName + "' does not exist.");
		}

		final List<String> result = new ArrayList<String>();

		// Run through the libs and put them into our list.
		final File[] files = this.assembleLibDir(projectName).listFiles();
		if (files != null) {
			for (final File file : files) {
				if (file.getName().endsWith("." + FSManager.LIB_EXTENSION)) {
					result.add(file.getName());
				}
			}
		}

		return result;
	}

	/**
	 * This method lists all available projects on the file system.
	 * 
	 * @return A list containing all available project names.
	 */
	public Collection<String> listAllProjects() {
		final List<String> result = new ArrayList<String>();

		// Get all directories within our root-dir
		final File[] files = new File(FSManager.ROOT_DIRECTORY).listFiles();
		for (final File file : files) {
			if (file.isDirectory()) {
				result.add(file.getName());
			}
		}

		return result;
	}

	/**
	 * This method copies the given source file to the given destination file.
	 * 
	 * @param src
	 *            The source element. This should be a file.
	 * @param dst
	 *            The destination element. This should be a (non existing) file.
	 * @throws IOException
	 *             If something went wrong. In this case the method will try to close all streams nevertheless.
	 */
	private void copyFile(final File src, final File dst) throws IOException {
		boolean result = true;
		// Open the files
		FileInputStream fileInputStream = null;
		FileOutputStream fileOutputStream = null;

		try {
			fileInputStream = new FileInputStream(src);
			fileOutputStream = new FileOutputStream(dst);
			final FileChannel inputChannel = fileInputStream.getChannel();
			final FileChannel outputChannel = fileOutputStream.getChannel();

			// Copy the data
			this.transfer(inputChannel, outputChannel, src.length());

			// This is just necessary to calm findbugs
			result = dst.setLastModified(src.lastModified());

		} catch (final IOException ex) {
			FSManager.LOG.error("An IO error occured", ex);
			result = false;
		} finally {
			// Try to close the streams
			if (fileInputStream != null) {
				try {
					fileInputStream.close();
				} catch (final IOException ex) {
					FSManager.LOG.error("An IO error occured", ex);
					result = false;
				}
			}
			if (fileOutputStream != null) {
				try {
					fileOutputStream.close();
				} catch (final IOException ex) {
					FSManager.LOG.error("An IO error occured", ex);
					result = false;
				}
			}
		}

		if (!result) {
			throw new IOException("An IO error occured");
		}
	}

	/**
	 * This method transfers data from one channel to the other.
	 * 
	 * @param fileChannel
	 *            The input channel.
	 * @param byteChannel
	 *            The output channel.
	 * @param lengthInBytes
	 *            The size in bytes to be copied.
	 * @throws IOException
	 *             If something went wrong.
	 */
	private void transfer(final FileChannel fileChannel, final ByteChannel byteChannel, final long lengthInBytes) throws IOException {
		long overallBytesTransfered = 0L;
		while (overallBytesTransfered < lengthInBytes) {
			final long count = Math.min(FSManager.BUF_SIZE_BYTES, lengthInBytes - overallBytesTransfered);
			final long bytesTransfered = fileChannel.transferTo(overallBytesTransfered, count, byteChannel);
			overallBytesTransfered += bytesTransfered;
		}
	}

	/**
	 * Checks whether a project with the name exists on the file system.
	 * 
	 * @param projectName
	 *            The name of the project.
	 * @return true if and only if a directory with the name of the project exists in the root dir.
	 */
	private boolean projectExists(final String projectName) {
		return this.assembleKaxFile(projectName).exists();
	}

	/**
	 * This method can be used to deliver the fully qualified URL of a given dependency for a given project.
	 * 
	 * @param lib
	 *            The library whose URL should be delivered.
	 * @param project
	 *            The corresponding project of the library.
	 * @return The URL to the given library if everything went well.
	 * @throws MalformedURLException
	 *             If the URL is for some reason invalid.
	 */
	public URL getURL(final MIDependency lib, final String project) throws MalformedURLException {
		final File file = new File(FSManager.ROOT_DIRECTORY + File.separator + project + File.separator + FSManager.LIB_DIRECTORY + File.separator
				+ lib.getFilePath());
		return file.toURL();
	}

	/**
	 * Assembles the {@link File}-element pointing to the directory of the given project.
	 * 
	 * @param projectName
	 *            The name of the project.
	 * @return The directory of the project.
	 */
	private File assembleProjectDir(final String projectName) {
		return new File(FSManager.ROOT_DIRECTORY + File.separator + projectName);
	}

	/**
	 * Assembles the {@link File}-element pointing to the kax-file of the given project.
	 * 
	 * @param projectName
	 *            The name of the project.
	 * @return The kax-file of the project.
	 */
	private File assembleKaxFile(final String projectName) {
		return new File(FSManager.ROOT_DIRECTORY + File.separator + projectName + File.separator + projectName + "." + FSManager.KAX_EXTENSION);
	}

	/**
	 * Assembles the {@link File}-element pointing to the library directory of the given project.
	 * 
	 * @param projectName
	 *            The name of the project.
	 * @return The library directory of the project.
	 */
	private File assembleLibDir(final String projectName) {
		return new File(FSManager.ROOT_DIRECTORY + File.separator + projectName + File.separator + FSManager.LIB_DIRECTORY);
	}

	/**
	 * Delivers the {@link URL}-element pointing to the kieker library.
	 * 
	 * @return The kieker library.
	 */
	public URL getKiekerURL() {
		return Thread.currentThread().getContextClassLoader().getResource(FSManager.KIEKER_LIB);
	}

	/**
	 * Delivers the kax-file for the given project.
	 * 
	 * @param projectName
	 *            The name of the project.
	 * @return The kax-file of the project.
	 */
	public File getProjectFile(final String projectName) {
		return this.assembleKaxFile(projectName);
	}

	/**
	 * This helper class is responsible for creating a classloader as a privileged action. This is recommended due to the java security manager.
	 * 
	 * @author Nils Christian Ehmke
	 * @version 1.0
	 */
	private static class PrivilegedClassLoaderAction implements PrivilegedAction<URLClassLoader> {
		/**
		 * The list of libraries used to create the class loader.
		 */
		private final List<URL> libs;

		/**
		 * Creates a new instance of this class using the given parameters.
		 * 
		 * @param libs
		 *            The list of libraries used to create the class loader.
		 */
		public PrivilegedClassLoaderAction(final List<URL> libs) {
			this.libs = libs;
		}

		/**
		 * Runs the action.
		 * 
		 * @return The class loader.
		 */
		@Override
		public URLClassLoader run() {
			// We use "null" as an explicit parent to make sure that the class loader is completely independent from the web gui class loader.
			return new URLClassLoader(this.libs.toArray(new URL[this.libs.size()]), null);
		}
	}

}
